use std::io;
use rgb::RGB8;

use crate::geometry::*;
use crate::task::Task;
use crate::time::Periodically;
use running_average::RealTimeRunningAverage;
use std::marker::PhantomData;
use std::fmt::Debug;

pub trait Shader: Send + Debug {
    fn draw(&self, surface_coords: &VirtualCoordinates, frame: usize) -> RGB8;
}

pub trait Surfaces<T: Surface>: Debug {
    fn new_surface(&mut self, area: &Rectangle<u8, Virtual>) -> Result<T, io::Error>;
}

pub trait Surface: Default + Clone + Debug {
    fn with_shader<F: FnMut(&dyn Shader)>(&self, f: F);
    fn set_shader(&mut self, shader: Box<dyn Shader>);
    fn clear_shader(&mut self);

    fn rect(&self) -> Rectangle<u8, Virtual>;
    fn set_rect(&mut self, rect: &Rectangle<u8, Virtual>);
    fn opacity(&self) -> u8;
    fn set_opacity(&mut self, opacity: u8);
}

pub trait Framed {
    fn start_frame(&mut self) {}
    fn end_frame(&mut self) {}
}

pub trait Display<T: Surface>: Surfaces<T> + Framed {
    fn render_frame(&mut self);
}

#[derive(Debug)]
pub struct Renderer<T: Display<S>, S: Surface> {
    display: T,
    fps: RealTimeRunningAverage<u32>,
    fps_display: Periodically,
    _sfc: PhantomData<S>
}

impl<T: Display<S>, S: Surface> Renderer<T, S> {
    pub fn new(display: T) -> Self {
        Self {
            display,
            fps: RealTimeRunningAverage::default(),
            fps_display: Periodically::new_every_n_seconds(5),
            _sfc: PhantomData::default()
        }
    }
}

impl<T: Display<S>, S: Surface> Task for Renderer<T, S> {
    fn name(&self) -> &'static str { "Renderer" }

    fn tick(&mut self) {
        self.display.start_frame();
        self.display.render_frame();
        self.display.end_frame();
        self.fps.insert(1);
        self.fps_display.run(|| {
            log::info!("FPS: {} {:?}", self.fps.measurement(), self.display);
        });
    }
}
